// originates from Ruetsch/Oster: Getting Started with CUDA
#include <cassert>
#include <stdlib.h>
#include <ctime>
#include <cblas.h>
#include <cublas.h>
#include <iostream>

using namespace std;

// const int BS=256; 
const int BS=64; 
// const int XL=1024;
// texture<int2> tex_u;

void CPU_matvec(double r[], const double M[], const double u[], const int n);
__global__ void GPU_matvec
      (double r[], const double M[], const double u[], const int n);
__global__ void GPU_matvec_T
      (double r[], const double M[], const double u[], const int n);
__global__ void GPU_matvec_stride
      (double r[], const double M[], const double u[], const int n);
__global__ void GPU_matvec_stride_0
      (double r[], const double M[], const double u[], const int n);
double compare(const double r_h[], const double r_cpu[], const int n);
double compare(double r_h[], const double r_d[], const double r_cpu[], const int n);

// rowise access
void CPU_matvec(double r[], const double M[], const double u[], const int n)
{
  for (int i=0; i<n; ++i)
    {
     double tmp = 0.0;
     int    i0  = i*n;
     for (int j=0; j<n; ++j)
       {
        tmp += M[i0+j] * u[j];
       }
     r[i] = tmp;
    }
}

// rowise access
__global__ void GPU_matvec
      (double r[], const double M[], const double u[], const int n)
{
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx < n)
    {
    double tmp = 0.0;
    int    i0  = idx*n;
    for (int j=0; j<n; ++j, ++i0)
       {
//         tmp += M[i0+j] * u[j];
        tmp += M[i0] * u[j];
       }
     r[idx] = tmp;
    }
}

//  columnwise access
__global__ void GPU_matvec_T
      (double r[], const double M[], const double u[], const int n)
{
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx < n)
    {
    double tmp = 0.0;
    int jn = idx;
    for (int j=0; j<n; ++j, jn += n)
       {
//         tmp += M[j*n+idx] * u[j];
        tmp += M[jn] * u[j];
       }
     r[idx] = tmp;
    }
}

// // columnwise access and u in texture
// __global__ void GPU_matvec_texture
//       (double r[], const double M[], const double u[], const int n)
// {
//   int idx = blockIdx.x * blockDim.x + threadIdx.x;
//   if (idx < n)
//     {
//     double tmp = 0.0;
//     for (int j=0; j<n; ++j)
//        {
//         const int2   c = tex1Dfetch(tex_u,j);
//         const double b = __hiloint2double(c.y,c.x);
//         tmp += M[j*n+idx] * b;
//        }
//      r[idx] = tmp;
//     }
// }

// rowise access  BS == 64 !!
__global__ void GPU_matvec_stride
      (double r[], const double M[], const double u[], const int n)
{
  __shared__ double sdata[BS];
  int tid = threadIdx.x;
  int idx = blockIdx.x * blockDim.x + tid;
  if (idx < n)
    {
      int i0= blockIdx.x * blockDim.x * n;
      for (int k=0; k<BS; ++k, i0+=n)
      {
       sdata[tid] = 0.0;
       for (int j=tid; j<n; j+=BS)
         {
          sdata[tid] += M[i0+j] * u[j];
         }
       __syncthreads();
//   Copyright: NVIDIA_CUDA_SDK4/projects/reduction/reduction_kernel.cu
//   Assumtion:  BS == 64  !!
        if (tid<32)  sdata[tid] += sdata[tid + 32];
	__syncthreads();
	if (tid<16)  sdata[tid] += sdata[tid + 16];
	__syncthreads();
	if (tid<8)   sdata[tid] += sdata[tid +  8]; 
	__syncthreads();
// 	if (tid<4)   sdata[tid] += sdata[tid +  4]; 
// 	__syncthreads();
// 	if (tid<2)  sdata[tid] += sdata[tid +  2]; 
// 	__syncthreads();
//         if (tid<1)  sdata[tid] += sdata[tid +  1]; 
// 	__syncthreads();
 
	 if (threadIdx.x==0)
	 {
           r[blockIdx.x * blockDim.x+k] = sdata[0]+sdata[1]+sdata[2]+sdata[3]+sdata[4]+sdata[5]+sdata[6]+sdata[7]; 
	 }
      __syncthreads(); 
      }
    }
}

// rowise access
__global__ void GPU_matvec_stride_0
      (double r[], const double M[], const double u[], const int n)
{
  __shared__ double tmp[BS];
  int tid = threadIdx.x;
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  if (idx < n)
    {
      int i0= blockIdx.x * blockDim.x * n;
      for (int k=0; k<BS; ++k, i0+=n)
      {
       tmp[tid] = M[i0+tid] * u[tid];
       for (int j=tid+BS; j<n; j+=BS)
         {
          tmp[tid] += M[i0+j] * u[j];
         }
       __syncthreads();

       if (tid==k)
	 {
	   double ac = tmp[0];
           for (int j=1; j<BS; ++j)  ac+=tmp[j];  // could be improved by cyclic reduction??
           r[blockIdx.x * blockDim.x+k] = ac;
	 }
       __syncthreads();
     }
    }
}


// // rowise access
// // shared memory xloc does'nt pay off - it slows down by factor 3
// __global__ void GPU_matvec_stride_2
//       (double r[], const double M[], const double u[], const int n)
// {
//   __shared__ double rloc[BS];
//   __shared__ double xloc[XL];
//   rloc[threadIdx.x] = 0.0;
//   int idx = blockIdx.x * blockDim.x + threadIdx.x;
//   __syncthreads();
//   if (idx < n)
//     {
//       for (int i0=0; i0<n; i0 += XL)
// 	{
// 	  //  block from global vector into shared vector
//           for (int k=threadIdx.x; k<XL; k+=BS)   xloc[k] = u[i0+k];
// 	  __syncthreads();    // Potential deadlock if (idx<n) ??
// 	  int ik = i0;
// 	  for (int k=0; k<blockDim.x; ++k, ik+=n)  //  row k
// 	  {
// 	  double tmp = 0.0;
// 	  for (int j=threadIdx.x; j<XL; j+=BS)
// 	    {
// 	      tmp += M[ik+j] * xloc[j];
// 	    }
// 	  rloc[k] += tmp;   // could be improved by cyclic reduction
// 	  }
// 	}
//     }
//   __syncthreads();
//   if (idx < n) r[idx] = rloc[threadIdx.x];
// }

//------------------------
double compare(const double r_h[], const double r_cpu[], const int n)
{
 double sum = 0.0;
 for (int i=0; i< n; ++i) sum += (r_cpu[i]-r_h[i])*(r_cpu[i]-r_h[i]);
 return sqrt(sum);
}
//------------------------
double compare(double r_h[], const double r_d[], const double r_cpu[], const int n)
{
 cudaMemcpy(r_h, r_d, n*sizeof(r_d[0]), cudaMemcpyDeviceToHost);
 
 return compare(r_h,r_cpu, n);
}
//------------------------

int main(void)
{
 double *A_h, *x_h, *r_h, *cpu_res; // host data
 double *A_d, *x_d, *r_d; // device data

 const int N = 64*30*5, LOOP = 10;
//  const int N = 10000, LOOP = 10;
//  const int N = 10240, LOOP = 10;;
 const size_t nBytes = N*sizeof(x_h[0]);

 const dim3 dimBlock (BS);
 const dim3 dimGrid( ceil( N / (double)BS) );

 double tgpu,tcpu;
 
 if (cublasInit() != CUBLAS_STATUS_SUCCESS) {
        fprintf(stderr, "ERROR: cublasInit() failed!\n");
        exit(1);
    }
 
 printf("%d rows in Matrix: %lu MByte Memory needed.\n", N, N*nBytes/1024/1024);  
 
 A_h = (double *)malloc(N*nBytes);
 x_h = (double *)malloc(nBytes);
 r_h = (double *)malloc(nBytes);
 cpu_res = (double *)malloc(nBytes);

 cudaMalloc((void **) &A_d, N*nBytes);
 cudaMalloc((void **) &x_d, nBytes);
 cudaMalloc((void **) &r_d, nBytes);

 for (int i=0; i<N; ++i) x_h[i] = (rand()+0.0)/RAND_MAX;
 for (int k=0; k<N; ++k)            // symmetric dense matrix, N*N entries
     for (int j=k; j<N; ++j)
          A_h[k*N+j] = A_h[j*N+k] = (rand()+0.0)/RAND_MAX;
//  uncomment next line for a non-symmetric matrix          
 for (int i=0; i<N*N; ++i) A_h[i] = (rand()+0.0)/RAND_MAX;

/* for (int i=0; i<N; ++i) x_h[i] = i+1;
 for (int i=0; i<N*N; ++i) A_h[i] = 0.0;
 for (int i=0; i<N*N; i+=N+1) A_h[i] = 4.0; */

 cudaMemcpy(A_d, A_h, N*nBytes, cudaMemcpyHostToDevice);
 cudaMemcpy(x_d, x_h, nBytes, cudaMemcpyHostToDevice);
 cudaMemcpy(r_d, r_h, nBytes, cudaMemcpyHostToDevice);
//  cudaMemcpy(b_d, a_d, nBytes, cudaMemcpyDeviceToDevice);

// #########################  CPU  #####################################
 auto tstart = clock();
 for (int k=0; k<LOOP; ++k)
     CPU_matvec(cpu_res, A_h, x_h, N);
 tcpu = static_cast<double>(clock() - tstart)/CLOCKS_PER_SEC;
//  printf("CPU  : %4.4f sec.\n", tcpu);
 printf("CPU      : %4.4f sec.   %4.1f GFLOPS  %4.1f GBps\n", tcpu, LOOP*2.0*N*N/tcpu/pow(2.0,30), LOOP*8*(1.0*N*N+2*N)/tcpu/pow(2.0,30)); 

// ######################### GPU #######################################
 tstart = clock();
 for (int k=0; k<LOOP*1; ++k)
     GPU_matvec<<<dimGrid, dimBlock>>>(r_d, A_d, x_d, N);
 cudaDeviceSynchronize();
 tgpu = static_cast<double>(clock() - tstart)/CLOCKS_PER_SEC;
 printf("GPU org  : %4.4f sec.   %4.1f GFLOPS  %4.1f GBps\n", tgpu, LOOP*2.0*N*N/tgpu*1e-9, LOOP*8*(1.0*N*N+N+N*dimGrid.x)/tgpu/pow(2.0,30));
 printf("             Error  : %g \n", compare(r_h, r_d, cpu_res, N) );

// ######################## GPU T  #####################################
 tstart = clock();
 for (int k=0; k<LOOP; ++k)
     GPU_matvec_T<<<dimGrid, dimBlock>>>(r_d, A_d, x_d, N);
 cudaDeviceSynchronize();
 tgpu = static_cast<double>(clock() - tstart)/CLOCKS_PER_SEC;
 printf("GPU T    : %4.4f sec.   %4.1f GFLOPS  %4.1f GBps\n", tgpu, LOOP*2.0*N*N/tgpu*1e-9, LOOP*8*(1.0*N*N+N+N*dimGrid.x)/tgpu/pow(2.0,30)); 
/*LOOP*8*(1.0*N*N+2*N)/tgpu/pow(2.0,30)); */
printf("             Error  : %g \n", compare(r_h, r_d, cpu_res, N) );
 
 // #######################  GPU stride 0  #############################
 tstart = clock();
 for (int k=0; k<LOOP; ++k)
     GPU_matvec_stride_0<<<dimGrid, dimBlock>>>(r_d, A_d, x_d, N);
 cudaDeviceSynchronize();
 tgpu = static_cast<double>(clock() - tstart)/CLOCKS_PER_SEC;
 printf("GPU str 0: %4.4f sec.   %4.1f GFLOPS  %4.1f GBps\n", tgpu, LOOP*2.0*N*N/tgpu*1e-9, LOOP*8*(1.0*N*N+N+N*dimGrid.x)/tgpu/pow(2.0,30));
 printf("             Error  : %g \n", compare(r_h, r_d, cpu_res, N) );
// #########################  GPU strode a  ############################
 tstart = clock();
 for (int k=0; k<LOOP; ++k)
     GPU_matvec_stride<<<dimGrid, dimBlock>>>(r_d, A_d, x_d, N);
 cudaDeviceSynchronize();
 tgpu = static_cast<double>(clock() - tstart)/CLOCKS_PER_SEC;
 printf("GPU str a: %4.4f sec.   %4.1f GFLOPS  %4.1f GBps\n", tgpu, LOOP*2.0*N*N/tgpu*1e-9, LOOP*8*(1.0*N*N+N+N*dimGrid.x)/tgpu/pow(2.0,30));
 printf("             Error  : %g \n", compare(r_h, r_d, cpu_res, N) );
 
 // ############################  CUBLAS  ##############################
 tstart = clock();
 for (int k=0; k<LOOP; ++k)
 {  // Attention: BLAS assumes columnwise storage as standard
     cublasDgemv('t', N, N, 1.0, A_d, N, x_d, 1, 0.0, r_d, 1);
         if (cublasGetError() != CUBLAS_STATUS_SUCCESS) {
            fprintf(stderr, "ERROR: cublasSgemv() failed!\n");
            exit(1);
        }
 }
 cudaDeviceSynchronize();
 tgpu = static_cast<double>(clock() - tstart)/CLOCKS_PER_SEC;
 printf("CUBLAS   : %4.4f sec.   %4.1f GFLOPS  %4.1f GBps\n", tgpu, LOOP*2.0*N*N/tgpu*1e-9,    LOOP*8*(1.0*N*N+N+N*dimGrid.x)/tgpu/pow(2.0,30)); 
 printf("             Error  : %g \n", compare(r_h, r_d, cpu_res, N) );

 free(r_h); free(x_h); cudaFree(r_d); cudaFree(x_d); free(cpu_res);
 free(A_h); cudaFree(A_d);
 return 0;
}
